{========================================================================}
{=                (c) 1995-99 SwiftSoft Ronald Dittrich                 =}
{========================================================================}
{=                          All Rights Reserved                         =}
{========================================================================}
{=  D 01099 Dresden             = Fax.: +49(0)351-8037944               =}
{=  Loewenstr.7a                = info@swiftsoft.de                     =}
{========================================================================}
{=  Actual versions on http://www.swiftsoft.de/index.html               =}
{========================================================================}
{=  This code is for reference purposes only and may not be copied or   =}
{=  distributed in any format electronic or otherwise except one copy   =}
{=  for backup purposes.                                                =}
{=                                                                      =}
{=  No Delphi Component Kit or Component individually or in a collection=}
{=  subclassed or otherwise from the code in this unit, or associated   =}
{=  .pas, .dfm, .dcu, .asm or .obj files may be sold or distributed     =}
{=  without express permission from SwiftSoft.                          =}
{=                                                                      =}
{=  For more licence informations please refer to the associated        =}
{=  HelpFile.                                                           =}
{========================================================================}
{=  $Date: 22.03.99 - 18:11:25 $                                        =}
{========================================================================}
unit MMCDDB;

{$I COMPILER.INC}

// If the Connection is closed from the user then a exception ocurs.

interface

uses
    Windows,
    SysUtils,
    ExtCtrls,
    Classes,
    Forms,
    ScktComp,
    MMUtils,
    MMString,
    MMObj,
    MMCDRead,
    MMCDATyp;

Type
    PMMCDDBSite  = ^TMMCDDBSite;
    TMMCDDBSite  = record
       Address   : string;
       Port      : integer;
       Location  : string;
    end;

    PMMCDDBMatch = ^TMMCDDBMatch;
    TMMCDDBMatch = record
       Category  : string;
       Title     : string;
       CDTitle   : string;
       Artist    : string;
       DiskID    : string
    end;

const
    def_ConnectTimeout = 30000;
    def_UseFirstMatch  = False;
    def_HostName       = '';
    def_DataRev        = 0;
    def_AppName        = '';
    def_CdTitle        = '';
    def_Artist         = '';
    def_Category       = '';
    def_Title          = '';
    def_Revision       = '1.0';
    def_Cdid           = '';
    def_Server         = 'cddb.cddb.com';
    def_Port           = 8880;
    def_Email          = '';

type
    EMMCDDBError        = class(Exception);

    {-- TMMCDDBSites ----------------------------------------------------------}
    TMMCDDBSites = class(TList)
    private
       function GetSite(Index: Integer): PMMCDDBSite;
    public
       destructor Destroy; override;

       procedure Clear;
       property  Items[Index: Integer]: PMMCDDBSite read GetSite; default;
    end;

    {-- TMMCDDBMatches --------------------------------------------------------}
    TMMCDDBMatches = class(TList)
    private
       function GetMatch(Index: Integer): PMMCDDBMatch;
    public
       destructor Destroy; override;

       procedure Clear;
       property  Items[Index: Integer]: PMMCDDBMatch read GetMatch; default;
    end;

    TMMCDDBErrorEvent   = procedure(Sender: TObject; Msg: string; var Handled: Boolean) of object;
    TMMCDDBSendData     = procedure(Sender: TObject; Msg: string) of object;
    TMMCDDBReceivedData = procedure(Sender: TObject; Msg: string) of object;
    TMMCDDBSelectMatch  = procedure(Sender: TObject; Matches: TMMCDDBMatches; var Match: integer) of object;
    TMMCDDBState        = procedure(Sender: TObject; Msg: string) of object;

    {-- TMMCDDB ---------------------------------------------------------------}
    TMMCDDB           = class(TMMNonVisualComponent)
    private
       FCDReader      : TMMCDReader;
       FClientSocket  : TClientSocket;
       FTimer         : TTimer;
       FWaiting       : Boolean; // true while waiting for a response
       FTerminator    : string;  // responce terminator
       FTimeout       : Boolean; // true if timer elapses before wait is over
       FServerResponce: string;  // latest responce
       FAbort         : Boolean;

       FConnectTimeout: Longint;
       FUseFirstMatch : Boolean;
       FServer        : string;
       FPort          : integer;
       FHostName      : string;
       FEmail         : string;
       FDataRev       : Integer;
       FAppName       : string;
       FTitle         : string;
       FCdTitle       : string;
       FArtist        : string;
       FCategory      : string;
       FRevision      : string;
       FCDID          : string;
       FQueryCategory : string;
       FQueryCDID     : string;
       FQueryTitle    : string;

       FSongs         : TStringList;
       FExtData       : TStringList;
       FCategories    : TStringList;
       FSites         : TMMCDDBSites;
       FMatches       : TMMCDDBMatches;

       FOnError       : TMMCDDBErrorEvent;
       FOnSendData    : TMMCDDBSendData;
       FOnReceiveData : TMMCDDBReceivedData;
       FOnSelectMatch : TMMCDDBSelectMatch;
       FOnState       : TMMCDDBState;

       function  cddbConnect: Boolean;
       function  cddbSend(SendString: string): integer;
       function  cddbGetCDID: Longint;
       function  cddbHello: Boolean;
       function  cddblscat: string;
       function  cddbQuery: string;
       procedure cddbQuit;
       function  cddbRead(ReadCategory, ReadCDID: string): string;
       function  cddbSites: string;
       function  cddbMotd: string;

       procedure CloseMatchDecode(Response: string);
       procedure QueryMatchDecode(Responce: string);
       function  PullData(var InputString: string; Header: string): string;
       function  ReadDecode(ReadString: string): string;

       procedure SocketRead(Sender: TObject; Socket: TCustomWinSocket);
       procedure TimeOutTimer(Sender: TObject);
       procedure SetCDReader(aValue: TMMCDReader);
       procedure SetConnectTimeout(aValue: integer);
       procedure SetHostName(aValue: string);
       procedure SetAppName(aValue: string);
       procedure SetRevision(aValue: string);
       procedure SetEmail(aValue: string);

    protected
       procedure Notification(AComponent: TComponent; Operation: TOperation); override;
       procedure Error(Msg: string); dynamic;
       procedure SendData(Msg: string); dynamic;
       procedure ReceiveData(Msg: string); dynamic;
       procedure State(Msg: string); dynamic;

    public
       constructor Create(aOwner: TComponent); override;
       destructor  Destroy; override;

       function  ReadCDDBSites: Integer;
       function  ReadCategories: integer;
       function  ReadMessageOfTheDay: string;
       function  ReadCDDB: string;

       procedure Clear;

       property Abort: Boolean read FAbort write FAbort;

       property CDID: Longint read cddbGetCDID;

       property DataRev: integer read FDataRev write FdataRev;

       property Title: string read FTitle write FTitle;
       property CDTitle: string read FCDTitle write FCDTitle;
       property Artist: string read FArtist write FArtist;
       property Category: string read FCategory write FCategory;

       property Songs : TStringList read FSongs;
       property ExtData: TStringList read FExtData;
       property Categories: TStringList read FCategories;
       property Sites: TMMCDDBSites read FSites;

    published
       property OnError: TMMCDDBErrorEvent read FOnError write FOnError;
       property OnSendData: TMMCDDBSendData read FOnSendData write FOnSendData;
       property OnReceiveData: TMMCDDBReceivedData read FOnReceiveData write FOnReceiveData;
       property OnSelectMatch: TMMCDDBSelectMatch read FOnSelectMatch write FOnSelectMatch;
       property OnState: TMMCDDBState read FOnState write FOnState;
       
       property ConnectTimeOut: integer read FConnectTimeOut write SetConnectTimeOut;
       property HostName: string read FHostName write SetHostName;
       property AppName: string read FAppName write SetAppName;
       property Revision: string read FRevision write SetRevision;
       property Email: string read FEmail write SetEmail;
       property Port: integer read FPort write FPort;
       property Server: string read FServer write FServer;
       property UseFirstMatch: Boolean read FUseFirstMatch write FUseFirstMatch;
       property CDReader: TMMCDreader read FCDReader write SetCDReader;
    end;

implementation

const
     CR_LF = #13#10;

{-------------------------------------------------------------------------}
function PosEx(Start: integer; SubStr, S: string): integer;
begin
   S := Copy(S,Start,Length(S));
   Result := Pos(SubStr,S);
end;

{------------------------------------------------------------------------------}
function NewLineFix(InputString: string): string;
var
   Location: integer;
begin
   { changes all "\n"s to CR_LF }
   Location := Pos('\n',InputString);
   while (Location > 0) do
   begin
      InputString := LeftStr(InputString, Location-1) + CR_LF + RightStr(InputString, Length(InputString)-Location-1);
      Location := Pos('\n',InputString);
   end;
   Result := InputString;
end;

{------------------------------------------------------------------------------}
function RemoveTopLine(InputString: string): string;
begin
   // removes top line in a multiline (contains vbCrLf's) string
   Result := RightStr(InputString, Length(InputString) - Pos(CR_LF,InputString)-1);
end;

{------------------------------------------------------------------------------}
function ResponceCode(Responce: string): integer;
begin
   Result := StrToIntDef(LeftStr(Responce, 3),0);
end;

{== TMMCDDBSites ==============================================================}
function TMMCDDBSites.GetSite(index: Integer): PMMCDDBSite;
begin
   Result:= inherited Items[index];
end;

{-- TMMCDDBSites --------------------------------------------------------------}
procedure TMMCDDBSites.Clear;
var
   i : integer;
begin
   for i := 0 to Count-1 do
       if Assigned(Items[I]) then Dispose(Items[I]);

   inherited Clear;
end;

{-- TMMCDDBSites --------------------------------------------------------------}
destructor TMMCDDBSites.Destroy;
begin
   Clear;

   inherited Destroy;
end;

{== TMMCDDBMatches ============================================================}
function TMMCDDBMatches.GetMatch(index: Integer): PMMCDDBMatch;
begin
   Result:= inherited Items[index];
end;

{-- TMMCDDBMatches ------------------------------------------------------------}
procedure TMMCDDBMatches.Clear;
var
   i : integer;
begin
   for i := 0 to Count-1 do
       if Assigned(Items[I]) then Dispose(Items[I]);

   inherited Clear;
end;

{-- TMMCDDBMatches ------------------------------------------------------------}
destructor TMMCDDBMatches.Destroy;
begin
   Clear;

   inherited Destroy;
end;

{== TMMCDDB ===================================================================}
constructor TMMCDDB.Create(aOwner: Tcomponent);
begin
   inherited Create(aOwner);

   FCdid          := def_Cdid;
   FServer        := def_Server;
   FPort          := def_Port;
   FEmail         := def_Email;
   FRevision      := def_Revision;
   FTitle         := def_Title;
   FCategory      := def_Category;
   FCdTitle       := def_CdTitle;
   FArtist        := def_Artist;
   FAppName       := def_AppName;
   FDataRev       := def_DataRev;
   FHostName      := def_HostName;
   FUseFirstMatch := def_UseFirstMatch;
   FConnectTimeout:= def_ConnectTimeout;
   FTerminator    := '';
   FCDReader      := nil;
   FAbort         := False;

   FClientSocket           := TClientsocket.Create(Self);
   FClientSocket.Active    := False;
   FClientSocket.ClientType:= ctNonBlocking;
   FClientsocket.OnRead    := SocketRead;

   FTimer                  := TTimer.Create(Self);
   FTimer.Enabled          := False;
   FTimer.Interval         := def_ConnectTimeout;
   FTimer.OnTimer          := TimeOutTimer;

   FSongs      := TStringList.Create;
   FExtData    := TStringList.Create;
   FCategories := TStringList.Create;
   FSites      := TMMCDDBSites.Create;
   FMatches    := TMMCDDBMatches.Create;
end;

{-- TMMCDDB -------------------------------------------------------------------}
destructor TMMCDDB.Destroy;
begin
   FClientSocket.Free;
   FTimer.Free;

   FSongs.Free;
   FExtData.Free;
   FCategories.Free;
   FSites.Free;
   FMatches.Free;

   inherited destroy;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.Error(Msg: string);
var
   Handled: Boolean;
begin
   Handled := False;

   if assigned(FOnError) then
      FOnError(Self,Msg,Handled);

   FClientSocket.Active := False;

   if not Handled then
      raise EMMCDDBError.Create(Msg);
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.State(Msg: string);
begin
   if assigned(FOnState) then
      FonState(Self,Msg);
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SendData(Msg: string);
begin
   if assigned(FOnSendData) then
   begin
      if Pos(CR_LF,Msg) > 0 then
         Msg := Copy(Msg,1,Pos(CR_LF,Msg)-1);
      FOnSendData(Self,Msg);
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.ReceiveData(Msg: string);
var
   S: string;
   i: integer;
begin
   if assigned(FOnReceiveData) then
   begin
      while Pos(CR_LF, Msg) > 0 do
      begin
         i := Pos(CR_LF,Msg);
         S := LeftStr(Msg,i-1);
         FOnReceiveData(Self,S);
         Msg := RightStr(Msg,Length(Msg)-i-1);
      end;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.Clear;
begin
   FSongs.Clear;
   FMatches.Clear;
   FExtData.Clear;
   FTitle    := '';
   FCDTitle  := '';
   FArtist   := '';
   FCategory := '';
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.Notification(AComponent: TComponent; Operation: TOperation);
begin
   inherited Notification(AComponent, Operation);

   if (Operation = opRemove) and (AComponent = FCDReader) then
      FCDReader := Nil;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetCDReader(aValue: TMMCDReader);
begin
   if Longint(aValue) = Longint(Self) then exit;

   if (aValue is TMMCDReader) or (aValue = nil) then
   begin
      FCDReader := aValue;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SocketRead(Sender: TObject; Socket: TCustomWinSocket);
var
   StrData: string;
begin
   if Socket.Connected then
   begin
      StrData := Socket.ReceiveText;

      if FTerminator <> '' then
      begin
         FServerResponce := FServerResponce + StrData;
         if RightStr(FServerResponce, 5) = CR_LF + '.' + CR_LF then
         begin
            ReceiveData(FServerResponce);
            FTerminator := '';
            FWaiting    := False;
         end;
      end
      else
      begin
         FServerResponce := StrData;
         ReceiveData(FServerResponce);
         FWaiting        := False;
      end;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.TimeOutTimer(Sender: TObject);
begin
   FClientSocket.Active := False;
   FTimeout := True;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbConnect: Boolean; // true means everything worked ok
var
  ConnectString: string;            // the string to send to the server

begin
   if not FClientSocket.Active then // it should be closed before we do this
   begin
      State('Finding Host...');

      if FServer = '' then
         FServer := def_Server;     // if the server is blank use the default server

      if FPort = 0 then
         FPort := def_Port;         // if the port is not set use the default port

      FClientSocket.Host   := FServer;
      FClientSocket.Port   := FPort;
      FClientSocket.Active := True; // open the connection

      ConnectString := 'Connecting to: ' + FServer + ' Port: ' + IntToStr(FPort) + CR_LF;
      SendData(ConnectString);

      FServerResponce := '';        // blank out the server responce

                                    // wait for responce
      FWaiting := True;             // waiting will be changed to false by the data arrival event of the
                                    // winsock control when data arrives
      FTimeout := False;            // reset the timeout
      FTimer.Enabled := True;       // start the timeout timer
      while FWaiting and not FTimeout and not FAbort do // 'wait until data has arrived or the timeout occurs
      begin
         Application.ProcessMessages;    // don't hog the computer with this loop
      end;

      FTimer.Enabled := False;          // turn of the timer
      if FTimeout or FAbort then        // if the timeout occured then exit the function
      begin
         if FAbort then FClientSocket.Active := False;
         Result := False;
      end
      else Result := True;
   end
   else
   begin
      SendData('Socket was already open');
      Result := True;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbSend(SendString: string): integer;
begin
   Result := 0;

   if not FClientSocket.Socket.Connected then cddbConnect;

   if (SendString <> '') and FClientSocket.Socket.Connected then
   begin
      FServerResponce := '';
      FWaiting := True;
      FTimeout := False;

      SendData(SendString);
      FClientSocket.Socket.SendText(SendString);

      FTimer.Enabled := True;
      while FWaiting and not FTimeout and not FAbort and not Application.Terminated do
      begin
         Application.ProcessMessages;
      end;
      FTimer.Enabled := False;
      if FTimeout then
      begin
         Result := -1;
      end
      else if FAbort then
      begin
         FClientSocket.Active := False;
         Result := -2;
      end
      else Result := 1;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbGetCDID: Longint;
type
  TMSFRec = record
    Minutes : byte;     {least significant byte}
    Seconds : byte;
    Frames  : byte;
    NotUsed : byte;     {most significant byte}
  end;

  function prg_sum(n: integer): integer;
  var
     buf: string;
     ib : Integer;
  begin
     buf    := IntToStr(n);
     Result := 0;
     for ib := 1 to Length(buf) do Result := Result + (StrToInt(Copy(Buf,ib,1)));
  end;

var
   i,N,L,SizeTotal: integer;

begin
   Result := 0;

   if (FCDReader = nil) then
       raise EMMCDDBError.Create('Not connected to CDReader');

   with FCDReader do
   begin
      if not DeviceReady then
         Error('Device is not ready');

      if not ReadTOC then
         Error('Unable to read CD directory');

      N := 0;
      L := 0;
      for i := 0 to Tracks.Count-1 do
      begin
         with TMSFRec(Tracks[i].Start) do
         begin
            N := N + prg_sum((minutes * 60) + seconds);
            L := L + MSFToFrame(Tracks[i].Length);
         end;
      end;

      with TMSFRec(FrameToMSF(L)) do
      begin
         L := minutes * 60 + seconds;
      end;
      Result := ((N mod $0FF) shl 24) xor (L shl 8) xor Tracks.Count;
      FCDID := IntToHex(Result,8);
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbHello: Boolean;
var
   HelloString: string;
begin
   Result := False;

   if (FEmail = '') then
      raise EMMCDDBError.Create('Need an email address');

   if (Pos('@',FEmail) <= 0) then
      raise EMMCDDBError.Create('Invalid email address');

   State('Logging in...');

   if (FHostName = '') then
       FHostName := RightStr(FEmail,Length(FEMail)-Pos('@',FEmail));

   HelloString := 'cddb hello ' + FEmail + ' ' + FHostName + ' ' + FAppName + ' ' + FRevision + CR_LF;
   Result := cddbSend(HelloString) = 1;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddblscat: string;
var
   LscatString: string;
   Res: integer;
begin
   Result      := '';
   LscatString := 'cddb lscat' + CR_LF;
   FTerminator  := '.';

   State('Asking for category list...');

   Res := cddbSend(lscatString);
   if Res = 0 then exit;

   if (Res = -1) then
   begin
      Result := '600 Timeout';
      exit;
   end
   else if (Res = -2) then
   begin
      Result := '700 Abort';
      exit;
   end;

   Result := FServerResponce;
   if LeftStr(FServerResponce, 3) = '210' then
   begin
      // remove responce code line
      Result := RightStr(FServerResponce, Length(FServerResponce) - Pos(CR_LF,FServerResponce)-1);
      while LeftStr(Result, 1) <> '.' do
      begin
         FCategories.Add(LeftStr(Result, Pos(CR_LF,Result)-1));
         Result := RightStr(Result, Length(Result) - Pos(CR_LF, Result)-1);
      end;
      Result := '210 Okay category list follows';
   end
   else
   begin
      Result := '401';
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbQuery: string;
var
   i,Res: integer;
   L: Longint;
   QueryString: string;
begin
   Result := '';

   if (FCDReader = nil) then exit;

   if QueryString = '' then
   with FCDReader do
   begin
      if FCdid = '' then // if there is no cdid yet get one
         cddbGetCDID;

      // query string is the the cdid followed the number of tracks, followed by the
      // start points of each song followed by the disc length
      QueryString := 'cddb query ' + FCdid + ' ' + IntToStr(FCDReader.Tracks.Count);
      L := 0;
      for i := 0 to Tracks.Count-1 do
      with Tracks[i] do
      begin
         QueryString := QueryString + ' ' + IntToStr(MSFToFrame(Start));
         L := L + MSFToFrame(Tracks[i].Length);
      end;
      QueryString := QueryString + ' ' + IntToStr(MSFToSecond(FrameToMSF(L))) + CR_LF;
   end;

   // TEST REMOVE ME - This generates a really long response
   // QueryString := 'cddb query 870a7b09 9 182 20392 44240 68915 97400 106915 125967 145350 167270 2683' + CR_LF;

   State('Querying database...');

   Res := cddbSend(QueryString);
   if Res = 0 then exit;

   if (Res = -1) then
   begin
      Result := '600 Timeout';
      exit;
   end
   else if (Res = -2) then
   begin
      Result := '700 Abort';
      exit;
   end;

   Result := FServerResponce;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.cddbQuit;
var
   SendString: string;
begin
   if FClientSocket.Socket.Connected then
   begin
      State('Disconnecting...');

      SendString := 'quit' + CR_LF;

      cddbSend(SendString);
      FServerResponce := '';
      FClientSocket.Active := False;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbRead(ReadCategory, ReadCDID: string): string;
var
   ReadString: string;
   Res: integer;
begin
   Result := '';
   if (ReadCDID = '') and (ReadCategory = '') then exit;

   ReadString := 'cddb read ' + ReadCategory + ' ' + ReadCDID + CR_LF;
   FTerminator := '.';

   State('Retrieving data...');

   Res := cddbSend(ReadString);
   if (Res = 0) then exit;

   if (Res = -1) then
       FServerResponce := '600 Timeout'
   else if (Res = -2) then
       FServerResponce := '700 Abort';

   Result := FServerResponce;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbSites: string;
var
   SiteString: string;
   i,Res: integer;
   Site: PMMCDDBSite;
begin
   Result := '';
   FSites.Clear;

   SiteString := 'sites' + CR_LF;
   FTerminator := '.';

   State('Asking for server list...');

   Res := cddbSend(SiteString);
   if (Res = 0) then exit;

   if (Res = -1) then
   begin
      Result := '600 Timeout';
      exit;
   end
   else if (Res = -2) then
   begin
      Result := '700 Abort';
      exit;
   end;

   Result := FServerResponce;
   case ResponceCode(FServerResponce) of
       210: begin // Ok, site information follows
                  // remove responce code line
               Result := RightStr(Result, Length(Result) - Pos(CR_LF,Result)-1);
               i := 0;
               while LeftStr(Result, 1) <> '.' do
               begin
                  New(Site);

                  Site.Address := LeftStr(Result, Pos(' ',Result)-1);
                  Result := RightStr(Result, Length(Result) - Length(Site.Address)-1);
                  Site.Port := StrToIntDef(LeftStr(Result, Pos(' ',Result)-1),0);
                  Result := RightStr(Result, Length(Result) - PosEx(2,' ',Result)-1);
                  Result := RightStr(Result, Length(Result) - PosEx(2,' ',Result)-1);
                  Result := RightStr(Result, Length(Result) - PosEx(2,' ',Result)-1);
                  Site.Location := LeftStr(Result, Pos(CR_LF, Result)-1);
                  Result := RightStr(Result, Length(Result) - Pos(CR_LF, Result)-1);

                  FSites.Add(Site);
               end;
               Result := '210 Ok, site information follows.'
            end;
       401: begin // No site information available.
               Result := '401 No site information available.'
            end;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.cddbMotd: string;
var
  MotdString: string;
  Res: integer;
begin
  Result := '';
  MotdString := 'motd' + CR_LF;
  FTerminator := '.';

  State('Asking for "MOD"...');

  Res := cddbSend(MotdString);
  if (Res = 0) then exit;

   if (Res = -1) then
   begin
      Result := '600 Timeout';
      exit;
   end
   else if (Res = -2) then
   begin
      Result := '700 Abort';
      exit;
   end;

  Result := FServerResponce;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.CloseMatchDecode(Response: string);
var
   S: string;
   Match: PMMCDDBMatch;
   Cat,Title,ID: string;
begin
   FMatches.Clear;
   // tear off top line
   Response := RightStr(Response, Length(Response) - Pos(CR_LF, Response)-1);
   while Length(Response) > 3 do
   begin

      S := LeftStr(Response, Pos(CR_LF, Response)-1);
      Cat := LeftStr(S, Pos(' ', S)-1);
      S := RightStr(S, Length(S) - Pos(' ', S));
      Id := LeftStr(S, Pos(' ', S)-1);
      S := RightStr(S, Length(S) - Pos(' ', S));
      Title := S;
      // tear off top line
      Response := RightStr(Response, Length(Response) - Pos(CR_LF, Response)-1);

      New(Match);
      Match.Category:= Cat;
      Match.DiskID  := ID;
      Match.Title   := Title;

      if Pos('/',Title) > 0 then
      begin
         Match.CDTitle := Trim(RightStr(Title, Length(Title)-Pos('/', Title)-1));
         Match.Artist  := Trim(LeftStr(Title, Pos('/',Title)-1));
      end
      else if Pos(' - ',Title) > 0 then
      begin
         Match.CDTitle := Trim(RightStr(Title, Length(Title)-Pos(' - ', Title)-1));
         Match.Artist  := Trim(LeftStr(Title, Pos(' - ',Title)-1));
      end
      else
      begin
         Match.CDTitle := Title;
         Match.Artist  := Title;
      end;

      FMatches.Add(Match);
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.QueryMatchDecode(Responce: string);
begin
   // trim up to first blank
   Responce := RightStr(Responce, Length(Responce) - Pos(' ', Responce));
   FQueryCategory := LeftStr(Responce, Pos(' ', Responce)-1);
   Responce := RightStr(Responce, Length(Responce) - Pos(' ', Responce));
   FQueryCDID := LeftStr(Responce, Pos(' ', Responce)-1);
   Responce := RightStr(Responce, Length(Responce) - Pos(' ', Responce));
   FQueryTitle := Responce;
   // code categ discid dtitle
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.PullData(var InputString: string; Header: string): string;
var
   Location: integer;
   DataLine: string;
begin
   // this function will pull all data out off a string that is after a header
   // even if it is on multiple lines
   Result := '';
   while Pos(Header,InputString) > 0 do // do while header is in string
   begin
      Location := Pos(Header, InputString); // find the location of the header
      InputString := RightStr(InputString, Length(InputString) - Location); // trim down to loaction found
      DataLine := LeftStr(InputString, Pos(CR_LF, InputString)-1); // get up to the CRLF
      DataLine := RightStr(DataLine, Length(DataLine)-Length(Header)+ 1); // remove the header
      Result := Result + DataLine; // add up pulled data
      InputString := RemoveTopLine(InputString); // romove top line for next go around
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.ReadDecode(ReadString: string): string;
var
   i: integer;
   ExtData: string;
begin
    if LeftStr(ReadString, 3) = '600' then // read timed out
    begin
       Result := '600 Timeout';
       exit;
    end
    else if LeftStr(ReadString, 3) = '700' then // user abort
    begin
       Result := '700 Timeout';
       exit;
    end;

    FCategory := Copy(ReadString, 5, PosEx(5, ' ', ReadString)-1);
    FDataRev  := StrToIntDef(PullData(ReadString, '# Revision:'),0);

    // Get title
    FTitle := PullData(ReadString, 'DTITLE=');

    // split title
    if Pos('/',FTitle) > 0 then
    begin
       FCDTitle := Trim(RightStr(FTitle, Length(FTitle)-Pos('/', FTitle)-1));
       FArtist  := Trim(LeftStr(FTitle, Pos('/',FTitle)-1));
    end
    else if Pos(' - ',FTitle) > 0 then
    begin
       FCDTitle := Trim(RightStr(FTitle, Length(FTitle)-Pos(' - ', FTitle)-1));
       FArtist  := Trim(LeftStr(FTitle, Pos(' - ',FTitle)-1));
    end
    else
    begin
        FCDTitle := FTitle;
        FArtist  := FTitle;
    end;

    // Get Song titles
    while Pos('TTITLE' + IntToStr(FSongs.Count) + '=', ReadString) > 0 do
    begin
       FSongs.Add(PullData(ReadString,'TTITLE' + IntToStr(FSongs.Count) + '='));
    end;

    // Get Extended Data
    ExtData := PullData(ReadString, 'EXTD=');
    while Pos('EXTT' + IntToStr(FextData.Count) + '=',ReadString) > 0 do
    begin
       FExtData.Add(PullData(ReadString,'EXTT' + IntToStr(FExtData.Count) + '='));
    end;

    Result := FTitle + CR_LF + IntToStr(FSongs.Count);
    for i := 0 to FSongs.Count-1 do
        Result := Result + CR_LF + Songs[i];
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.ReadCDDB: string;
var
  QueryResponce: string;
  Match: integer;
begin
   Result := '';
   FAbort := False;

   Clear;

   cddbGetCDID;
   if cddbConnect then
   begin
      try
         if cddbHello then
         begin
            QueryResponce := cddbQuery;
            case ResponceCode(QueryResponce) of
               200: begin
                       // Found Exact Match
                       QueryMatchDecode(QueryResponce);
                       Result := ReadDecode(cddbRead(FQueryCategory, FQueryCDID));
                    end;
               202: begin
                       Result := '202 No exact match';
                       Error(Result);
                    end;
               211: begin
                       // Found inexact matches, list follows (until terminating marker)
                       CloseMatchDecode(QueryResponce);
                       if (FMatches.Count > 0) then
                       begin
                          if not FUseFirstMatch and assigned(FOnSelectMatch) then
                          begin
                             Match := -1;
                             FOnSelectMatch(Self,FMatches,Match);
                          end
                          else Match := 0;

                          if (Match >= 0) then
                          begin
                             FQueryCDID     := FMatches[Match].DiskId;
                             FQueryCategory := FMatches[Match].Category;
                             FQueryTitle    := FMatches[Match].Title;
                             Result := ReadDecode(cddbRead(FQueryCategory, FQueryCDID));
                             FMatches.Clear;
                          end;
                       end
                       else Result := '202 No exact match';
                    end;
               403: begin
                       //  Database entry Is corrupt
                       Result := '403 Database entry Is corrupt';
                       Error(Result);
                    end;
               409: begin
                       // No handshake
                       Result := '409 No handshake';
                       Error(Result);
                     end;
                600: begin
                        // timed out
                        Error('Timeout');
                     end;
                700: begin
                        // user abort
                        exit;
                     end;
            end;
         end
         else if FAbort then
         begin
            Result := '700 User Abort';
         end
         else
         begin
            Result := '431 Handshake not successful, closing connection';
            Error(Result);
         end;

      finally
         cddbQuit;
      end;
   end
   else if not FAbort then Error('000 Failure to Connect');
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.ReadCDDBSites: Integer;
begin
   Result := 0;
   FAbort := False;
   FSites.Clear;

   if cddbConnect then
   begin
      if cddbHello then
      begin
         if LeftStr(cddbSites, 3) = '210' then Result := Sites.Count;
         cddbQuit;
      end
      else if not FAbort then
      begin
         Error('431 Handshake not successful, closing connection');
      end;
   end
   else if not FAbort then Error('000 Failure to Connect');
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.ReadMessageOfTheDay: string;
var
   S: string;
begin
   Result := 'Error';
   FAbort := False;

   if cddbConnect then
   begin
      if cddbHello then
      begin
         S := cddbMotd;
         if LeftStr(S, 3) = '210' then
            Result := S;
         cddbQuit;
      end
      else if not FAbort then
      begin
         Error('431 Handshake not successful, closing connection');
      end;
   end
   else if not FAbort then Error('000 Failure to Connect');
end;

{-- TMMCDDB -------------------------------------------------------------------}
function TMMCDDB.ReadCategories: integer;
begin
   Result := 0;
   FAbort := False;

   if cddbConnect then
   begin
      if cddbHello then
      begin
         if LeftStr(cddblscat, 3) = '210' then Result := FCategories.Count;
         cddbQuit;
      end
      else if not FAbort then
      begin
         Error('431 Handshake not successful, closing connection');
      end;
   end
   else if not FAbort then Error('000 Failure to Connect');
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetConnectTimeout(aValue: integer);
begin
   aValue := MinMax(aValue,5000,65535);
   if (aValue <> FConnectTimeOut) then
   begin
      FConnectTimeout := aValue;
      FTimer.Interval := aValue;
   end;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetHostName(aValue: string);
begin
   if Pos(' ', aValue) > 0 then
      Error('No spaces allowed in HostName');
   FHostName := aValue;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetAppName(aValue: string);
begin
   if Pos(' ', aValue) > 0 then
      Error('No spaces allowed in AppName');
   FAppName := aValue;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetRevision(aValue: string);
begin
   if Pos(' ', aValue) > 0 then
      Error('No spaces allowed in Revision');
   FRevision := aValue;
end;

{-- TMMCDDB -------------------------------------------------------------------}
procedure TMMCDDB.SetEmail(aValue: string);
begin
   if Pos(' ',aValue) > 0 then
      Error('No spaces allowed in Email');

   if Pos('@',aValue) <= 0 then
      Error('Invalid Email address');

   FEmail := aValue;
end;

end.

